%% gh -- a compiler for Generic Haskell.
%% Copyright (c) 2001  The Generic Haskell Team. Utrecht University
%% 
%% This library is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Lesser General Public License as
%% published by the Free Software Foundation; either version 2.1 of the
%% License, or (at your option) any later version.
%% 
%% This library is distributed in the hope that it will be useful, but
%% WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.
%% 
%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
%%
%% $Id: ErrorMessages.lhs,v 1.2 2002/04/15 11:56:00 cvs-4 Exp $
%%
%% author: Andres Loeh (andres@cs.uu.nl)

%***************************************************************************
\section{Error messages}
%***************************************************************************

\begin{code}
module ErrorMessages where

import ErrorUtils
import AbstractD
import Pretty
import qualified PrettyPrint as PP
\end{code}

Data type for error messages.

\begin{code}

data GHError = GHError_TypeMError  QualName Position String Position String
             | GHError_TypeUError  Position PP_Doc String
             | GHError_ConstrMError  QualName Position Position
             | GHError_ConstrUError  Position PP_Doc
             | GHError_ValueMError QualName Position String Position String
             | GHError_ValueUError Position PP_Doc String
             | GHError_TopLevelOnlyError QualName Position String
             | GHError_ResolveError Range QualName
             | GHError_NameClashError Range QualName Entities
             | GHError_PolyBindingMError QualName QualName Position Position
             | GHError_NoTypeSigError QualName Position
             | GHError_NoPolyBindingWarn QualName Position
             | GHError_PolyBindingWellError QualName Position
             | GHError_PolyWellError QualName Position
             | GHError_KindUAError Position PP_Doc PP_Doc Kind Kind
             | GHError_KindUTError Position PP_Doc String Kind Kind
             | GHError_KindUCError Position PP_Doc (Maybe PP_Doc) Kind
             | GHError_KindUTsError Position PP_Doc Kind
             | GHError_KindUnkIError Position PP_Doc
             | GHError_KindUnkTWarn  Position PP_Doc
             | GHError_RequestAbstractError QualName Type
             | GHError_RequestNonTerminate Int

isError :: GHError -> Bool
isError (GHError_NoPolyBindingWarn _ _) = False
isError (GHError_KindUnkTWarn _ _) = False
isError _ = True

printErrorV :: Int -> GHError -> String
printErrorV v e | isError e = printError e
                | v > 0     = printError e
                | otherwise = ""

printError :: GHError -> String
printError (GHError_TypeMError name sl1 t1 sl2 t2)
  = errorLayout ( show sl1 ++ " and " ++ show sl2 
               ++ " ERROR: NAME CLASH ON TYPE LEVEL\n"
               ++ t1 ++ " and " ++ t2 ++ " at given locations both have name "
                     ++ (pretty $ PP.ppQualName False name) ++ "\n"
                )
printError (GHError_TypeUError sl name t)
  = errorLayout ( show sl ++ " ERROR: UNKNOWN TYPE IDENTIFIER\n"
               ++ pretty name ++ " is not a known " ++ t ++ "\n"
                )
printError (GHError_ConstrMError name sl1 sl2)
  = errorLayout ( show sl1 ++ " and " ++ show sl2 ++ " ERROR: NAME CLASH FOR CONSTRUCTORS\n"
               ++ "Two constructors have name "
                     ++ (pretty $ PP.ppQualName False name) ++ "\n"
                )
printError (GHError_ConstrUError sl name)
  = errorLayout ( show sl ++ " ERROR: UNKNOWN CONSTRUCTOR\n"
               ++ pretty name ++ " is not a known constructor name\n"
                )
printError (GHError_ValueMError name sl1 t1 sl2 t2)
  = errorLayout ( show sl1 ++ " and " ++ show sl2 
               ++ " ERROR: NAME CLASH ON VALUE LEVEL\n"
               ++ t1 ++ " and " ++ t2 ++ " at given locations both have name "
                     ++ (pretty $ PP.ppQualName False name) ++ "\n"
                )
printError (GHError_ValueUError sl name t)
  = errorLayout ( show sl ++ " ERROR: UNKNOWN TYPE IDENTIFIER\n"
               ++ pretty name ++ " is not a known " ++ t ++ "\n"
                )
printError (GHError_TopLevelOnlyError name sl t)
  = errorLayout ( show sl ++ " ERROR: TOPLEVEL-ONLY STATEMENT IN LOCAL POSITION\n"
               ++ t ++ " " ++ (pretty $ PP.ppQualName False name) ++ " has been declared local,\n"
               ++ "but this is not (yet) allowed\n"
                )
printError (GHError_ResolveError r name)
  = errorLayout ( show r ++ " ERROR: UNRESOLVED NAME\n"
               ++ (pretty $ PP.ppQualName False name) ++ " is unknown\n"
                )
printError (GHError_NameClashError r name entities)
  = errorLayout ( show r ++ " ERROR: NAME CLASH\n"
               ++ (pretty $ PP.ppQualName False name) ++ " refers to entities " 
               ++ errorEntities entities ++ "\n"
                )
printError (GHError_PolyBindingMError name st sl1 sl2)
  = errorLayout ( show sl1 ++ " and " ++ show sl2 
               ++ " ERROR: MULTIPLY DEFINED GENERIC FUNCTION BINDING\n"
               ++ "for type " ++ (pretty $ PP.ppQualName False st) ++ " and function "
                     ++ (pretty $ PP.ppQualName False name) ++ "\n"
                )
printError (GHError_NoTypeSigError name sl)
  = errorLayout ( show sl ++ " ERROR: TYPE SIGNATURE MISSING\n"
               ++ "for generic function " ++ (pretty $ PP.ppQualName False name) ++ "\n"
                )
printError (GHError_NoPolyBindingWarn name sl)
  = errorLayout ( show sl ++ " WARNING: SIGNATURE WITHOUT DEFINITION\n"
               ++ "for generic function " ++ (pretty $ PP.ppQualName False name) ++ "\n"
                )
printError (GHError_PolyBindingWellError name sl)
  = errorLayout ( show sl ++ " ERROR: MALFORMED GENERIC FUNCTION\n"
               ++ "generic function " ++ (pretty $ PP.ppQualName False name)
               ++ " has a variable case at given location,\n"
               ++ "but has other cases.\n"
                )
printError (GHError_PolyWellError name sl)
  = errorLayout ( show sl ++ " ERROR: MALFORMED KIND-INDEXED TYPE\n"
               ++ "type " ++ (pretty $ PP.ppQualName False name)
               ++ " needs exactly two cases, one for *, one for (->)\n"
                )
printError (GHError_KindUAError sl p1 p2 k1 k2)
  = errorLayout ( show sl ++ " ERROR: KIND UNIFICATION FAILED\n"
               ++ "application of " ++ pretty p1 ++ " :: " ++ (pretty $ PP.ppKind False k1) ++ "\n"
               ++ "            to " ++ pretty p2 ++ " :: " ++ (pretty $ PP.ppKind False k2) ++ "\n"
                )
printError (GHError_KindUTError sl name d ke ki)
  = errorLayout ( show sl ++ " ERROR: KIND UNIFICATION FAILED\n"
               ++ "for " ++ d ++ " " ++ (pretty $ name) ++ "\n"
               ++ "expected kind " ++ (pretty $ PP.ppKind False ke) ++ "\n"
               ++ "inferred kind " ++ (pretty $ PP.ppKind False ki) ++ "\n"
                )
printError (GHError_KindUCError sl name fname k)
  = errorLayout ( show sl ++ " ERROR: CONSTRUCTOR FIELD HAS NOT KIND *\n"
               ++ "inferred kind " ++ (pretty $ PP.ppKind False k) 
               ++ " for " ++ ff fname ++ " in constructor " ++ (pretty $ name) ++ "\n"
                )
 where ff Nothing  = "unlabelled field"
       ff (Just f) = "field " ++ (pretty f)
printError (GHError_KindUTsError sl name k)
  = errorLayout ( show sl ++ " ERROR: TYPE SYNONYM RHS HAS NOT KIND * \n"
               ++ "inferred kind " ++ (pretty $ PP.ppKind False k)
               ++ " for rhs in declaration of type " ++ (pretty $ name) ++ "\n"
                )
printError (GHError_KindUnkIError sl name)
  = errorLayout ( show sl ++ " ERROR: UNKNOWN IDENTIFIER DURING KIND INFERENCE\n"
               ++ "identifier " ++ pretty name ++ " not in scope\n"
                )
printError (GHError_KindUnkTWarn  sl name)
  = errorLayout ( show sl ++ " WARNING: UNKNOWN TYPE DURING KIND INFERENCE\n"
               ++ "type " ++ pretty name ++ " unknown\n"
                )

printError (GHError_RequestAbstractError name tp)
  = errorLayout ("ERROR: SPECIALISATION REQUEST FOR ABSTRACT TYPE\n"
               ++ "there is a specialisation request for " ++ (pretty $ PP.ppQualName False name) ++ "\n"
               ++ "                                   to " ++ (pretty $ PP.ppType False tp) ++ "\n"
               ++ "which is an abstract type and has no special case\n"
               ++ "in the function definition.\n"
                )

printError (GHError_RequestNonTerminate n)
  = errorLayout ("ERROR: SPECIALISATION REQUEST COMPUTATION DID NOT TERMINATE\n"
               ++ "after " ++ show n ++ " iterations.\n"
               ++ "You can try to increase the number of iterations with the\n"
               ++ "--cut command line option.\n"
                )

errorEntities :: [Entity] -> String
errorEntities []     = ""
errorEntities [e]    = errorEntity e
errorEntities (e:es) = errorEntity e ++ " and " ++ errorEntities es

errorEntity :: Entity -> String
errorEntity (Entity_Entity Nothing n)  = (pretty $ PP.ppName False n)
errorEntity (Entity_Entity (Just m) n) = m ++ "." ++ (pretty $ PP.ppName False n)

\end{code}

