%% gh -- a compiler for Generic Haskell.
%% Copyright (c) 2001  The Generic Haskell Team. Utrecht University
%% 
%% This library is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Lesser General Public License as
%% published by the Free Software Foundation; either version 2.1 of the
%% License, or (at your option) any later version.
%% 
%% This library is distributed in the hope that it will be useful, but
%% WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.
%% 
%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
%%
%% $Id: KindUtils.lhs,v 1.2 2002/04/15 11:56:01 cvs-4 Exp $
%%
%% author: Andres Loeh (andres@cs.uu.nl)

%***************************************************************************
\section{Kind inference utilities}
%***************************************************************************

\begin{code}
module KindUtils where

import FMap
import AbstractD
import qualified PrettyPrint as PP
import Pretty
import Maybe
\end{code}

%---------------------------------------------------------------------------
\subsection{Some simple functions}
%---------------------------------------------------------------------------

Create a simple chain of Star applications.

\begin{code}
simpleKind                      :: Int -> Kind
simpleKind 0                    =  Kind_Star
simpleKind n                    =  Kind_Arr Kind_Star (simpleKind (n-1))
\end{code}

Compute the rank of a kind. 

\begin{code}
rank                            :: Kind -> Int
rank Kind_Star                  =  1
rank (Kind_Arr l r)             =  max (1 + rank l) (rank r)
\end{code}

Determine the number of arguments of a kind.

\begin{code}
kindargs                        :: Kind -> Int
kindargs Kind_Star              =  0
kindargs (Kind_Arr l r)         =  1 + kindargs r
\end{code}

Check whether a given kind is a kind variable. A generic function
is called \emph{restricted} iff the kind of its type variable is
not a kind variable.

\begin{code}
isKindVar                       :: Kind -> Bool
isKindVar (Kind_Var _)          =  True
isKindVar _                     =  False

isRestricted                    :: Kind -> Bool
isRestricted                    =  not . isKindVar
\end{code}

Finally, we need the function to remove all remaining variables from 
a kind. This is achieved by transforming a given substitution.

\begin{code}
removeKindVars                  :: Kind -> KSubst -> KSubst
removeKindVars k s              =  rkv (applyKind s k) s
  where rkv Kind_Star      s    =  s
        rkv (Kind_Arr l r) s    =  rkv r (rkv l s)
        rkv (Kind_Var v)   s    =  add s (v,Kind_Star)
        rkv Kind_Error     s    =  s
\end{code}

%---------------------------------------------------------------------------
\subsection{Kind unification}
%---------------------------------------------------------------------------

Unification of two kinds is always relative to a current substitution.
The substitution is changed so that applied to both given kinds it will
result in their most general unifier. If that is not possible, |Nothing|
is returned.

\begin{code}
type KSubst                     =  FMap QualName Kind

unify                           :: Kind -> Kind  -- two input kinds
                                -> KSubst -> (KSubst,Bool)
unify k1 k2 s                   =  unify' (applyKind s k1) (applyKind s k2) s
  where
\end{code}

If one kind is an error kind, then unification succeeds without changes.

\begin{code}
  unify' Kind_Error _        s  =  (s,True)
  unify' _        Kind_Error s  =  (s,True)
\end{code}

If both kinds are |Kind_Star|, then we succeed without doing something.

\begin{code}
  unify' Kind_Star Kind_Star s  =  (s,True)
\end{code}

If one of both kinds is a variable, then we bind the variable to
the other kind.

\begin{code}
  unify' (Kind_Var v) k s       =  varBind v k s
  unify' k (Kind_Var v) s       =  varBind v k s
\end{code}

In the case of two arrow kinds we first unify the left hand sides,
then unify the right hand sides making use of the knowledge we have
gained from the unification of the left hand sides. In the end, the
composition of both substitutions is returned.

\begin{code}
  unify' (Kind_Arr l1 r1)
         (Kind_Arr l2 r2) s     =  let (s1,p ) = unify l1 l2 s
                                       (s2,p') = unify r1 r2 s1
                                   in  (s2,p && p')
\end{code}

All other kinds cannot be unified.

\begin{code}
  unify' _ _ s                  =  (s,False)  -- !!
\end{code}

A simple version of |unify| just checks whether two kinds could be
unified.

\begin{code}
unifyable                       :: Kind -> Kind -> Bool
unifyable k1 k2                 =  snd $ unify k1 k2 emptyFM
\end{code}

Binding a variable to a kind in a substitution.

\begin{code}
varBind                         :: QualName -> Kind -> KSubst -> (KSubst,Bool)
varBind v k@(Kind_Var v') s
  | v == v'                     =  (s,True)   -- nothing to do
  | otherwise                   =  (add s (v,k),True)
varBind v k s
  | v `elemKV` k                =  (s,False)  -- this is not good
  | otherwise                   =  (add s (v,k),True)
\end{code}

The function elemKV checks for occurences of a kind variable in a
kind.

\begin{code}
elemKV                          :: QualName -> Kind -> Bool
elemKV v (Kind_Arr l r)         =  elemKV v l || elemKV v r
elemKV v (Kind_Star)            =  False
elemKV v (Kind_Var v')          =  v == v'
elemKV v (Kind_Error)           =  False
\end{code}

Apply a substitution to a kind.

\begin{code}
applyKind                       :: KSubst -> Kind -> Kind
applyKind subst Kind_Star       =  Kind_Star
applyKind subst (Kind_Arr l r)  =  Kind_Arr (applyKind subst l) 
				            (applyKind subst r)
applyKind subst (Kind_Var v)    =  case find subst (v) of
                                     Just x -> applyKind subst x
                                     _      -> Kind_Var v
applyKind subst Kind_Error      =  Kind_Error
\end{code}

Do the same, but without recursive applications.

\begin{code}
substKind                       :: KSubst -> Kind -> Kind
substKind subst Kind_Star       =  Kind_Star
substKind subst (Kind_Arr l r)  =  Kind_Arr (substKind subst l)
                                            (substKind subst r)
substKind subst (Kind_Var v)    =  case find subst v of
                                     Just x -> x
                                     _      -> Kind_Var v
substKind subst Kind_Error      =  Kind_Error
\end{code}

Compose two kind substitutions.

\begin{code}
compKind                        :: KSubst -> KSubst -> KSubst
compKind sub1 sub2              =  (sub1 `thenFM` sub2) applyKind
\end{code}

Show a kind substitution (nicely).

\begin{code}
ppKSubst                        :: String {- Modid -} -> KSubst -> String
ppKSubst m k                    =  unlines
                                .  map pretty
                                .  map (\(x,y) -> PP.ppQualName True x >#< indent' x "::" 
                                                                       >#< PP.ppKind True y
                                       )
                                .  filter (\(x,_) -> isFromModule m x)
                                .  toList
                                $  k
  where indent' x               =  indent ((30 - length (pretty $ PP.ppQualName True x)) `max` 0)
        isFromModule m (QualName_Entity _ (Entity_Entity (Just m') n) _) = m == m'
        isFromModule _ _                                                 = False
\end{code}
