%% gh -- a compiler for Generic Haskell.
%% Copyright (c) 2001  The Generic Haskell Team. Utrecht University
%% 
%% This library is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Lesser General Public License as
%% published by the Free Software Foundation; either version 2.1 of the
%% License, or (at your option) any later version.
%% 
%% This library is distributed in the hope that it will be useful, but
%% WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.
%% 
%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
%%
%% $Id: Main.lhs,v 1.13 2002/07/05 14:44:14 cvs-4 Exp $
%%
%% author: andres@cs.uu.nl

%***************************************************************************
\section{Acknowledgements}
%***************************************************************************

Jan de Wit has written a prototype of the Generic Haskell compiler.

%***************************************************************************
\section{Main program}
%***************************************************************************

%---------------------------------------------------------------------------
\subsection{About the structure}
%---------------------------------------------------------------------------

Currently, the Generic Haskell compiler consists of native Haskell
modules as well as attribute grammar (\AG) files and grammar
descriptions for the frown parser generator. 

You can identify the type of the file by its extension:

\begin{center}
\begin{tabular}{ll} \hline 
|.hs|  & Haskell \\
|.lhs| & literate Haskell \\
|.lag| & literate \AG \\
|.inc.lag| & literate \AG include \\
|.lg|  & literate frown grammar \\ \hline
\end{tabular}
\end{center}

\AG source files are feeded into the |UU_AG| system to produce Haskell
files (which have the same name, but the extension |.hs|). \AG include
files are not standalone \AG source files, but are included into other
\AG files. Thus, there are no Haskell files of the same name.

Frown source files contain grammar descriptions that are used to
generate a parser. The generated Haskell file has again the same name
as the grammar file, but the extension |.hs|.

The file |Main.lhs| is the central Haskell module. It handles
the processing of command line options and the calls different
phases of the compiler. 

These are:
\begin{itemize}
\item parsing of the input file (|GHParser_NG|),
\item handling of interface files and imports (|Imports|),
\item static error analysis (|Analysis|),
\item compiler and output code generator (|GH|).
\end{itemize}

\paragraph{Compiling the program} This module is the main
module of the project. It should be possible to use automated
make-utilities such as |hmake|, the GHC make option (since
GHC 5.00), or the hugs import chaser, from this module to
get the entire project compiled.

\paragraph{Running in hugs/ghci} If you want to run Generic
Haskell out of an interactive environment, then |findAndCompile|
is the function to call from the interactive command line.

\paragraph{More information} For more information about how
to use the \GH\ compiler, please look at the \GH User's Manual.

\begin{code}

module Main where

\end{code}

%---------------------------------------------------------------------------
\subsection{Auxiliary Haskell modules}
%---------------------------------------------------------------------------

We need a few standard Haskell libraries (or libraries that are
in |hslibs|).

\begin{code}
import System                   -- these are for option handling
import GetOpt                   -- etc.
import Maybe
import List hiding (find)
import Monad
import IO
-- import IOExts                   {- for debugging -}
\end{code}

Then there are a few basic utility libraries that are part of
Generic Haskell (or are, at least, distributed with Generic
Haskell).

\begin{code}
import FileNameUtils            {- basic file handling -}
import Name                     {- names of generated functions -}
import Utils                    {- general utilities -}
import Pretty                   {- wrapper around pretty printing library -}
import OSSpecific
\end{code}

Sometimes, the need arises to show debugging output even in
|Main|. In these cases, it comes handy to import the
module that pretty prints the abstract syntax of Generic
Haskell.

\begin{code}
-- temporary!!
-- import PrettyPrint as PP
\end{code}

%---------------------------------------------------------------------------
\subsection{Compiler phases}
%---------------------------------------------------------------------------

We need the abstract grammar with all extensions, and the parser.

\begin{code}
import AbstractD
import AbstractPP
import qualified GHParser_NG as GHP
\end{code}

We need |Imports.hs| to deal with imported modules, 
|NameReplace.hs| to resolve all names,
|Analysis.hs| to do static error analysis.
We then import |GH.hs| that is supposed to deal with almost everything.

\begin{code}
import qualified Imports as I
import qualified NameReplace as NR
import qualified Analysis as A
import qualified GH as GH
\end{code}

%---------------------------------------------------------------------------
\subsection{Compilation}
%---------------------------------------------------------------------------

The function |findAndCompile| determines the correct input and
output file name for the call to the compiler. The extension defaults
to |ghs|, but can be different. The output file is automatically the
same basename with extension |hs|. The list of flags contains the
command line options. The corresponding data type is described near
the end of this file.

\begin{code}

findAndCompile                  :: [Flag] -> String -> IO ()
findAndCompile opts name        =  
  let make                      =  not . null . filter (==Make) $ opts
      continue                  =  if make then (\c -> determineDependencies True opts [] c >> return ())
                                           else (\(_,g,i,r) -> compile opts g i r)
  in  case (extension name) of 
        Nothing                 -> continue (name
                                            ,name ++ ".ghs"
                                            ,name ++ ".ghi" 
                                            ,name ++  ".hs")
        (Just "hs")             -> error "Source file is Haskell file."
        (Just ext)              -> 
          let bname = take (length name - (length ext + 1)) name
          in  continue (bname,name
                       ,bname ++ ".ghi"
                       ,bname ++  ".hs")

\end{code}

The function |determineDependencies| issues the correct sequence
of |compile| commands to compile a file with dependencies.

\begin{code}

determineDependencies           :: Bool -> [Flag] -> [Call] -> Call -> IO Bool
determineDependencies top opts known from@(frommod,fromghs,fromghi,fromhs)
                                =  do
    -- copied from compile
    let verbosity               =  length (filter (==Verbose) opts)
    sp                          <- if top then expandPath $
                                               foldr (\o r -> 
                                                     case o of
                                                       Path p -> p : r
                                                       _      -> r) 
                                                     searchPath opts
                                          else return []
    let atVerbosity n s         =  when (verbosity >= n) $ putStrLn s 

    -- Parse input file
    (root',fn)                  <- (if top then id else (\a -> catch a (\ _ -> return (Nothing,fromghs))))
                                     $ GHP.parseFile (verbosity > 0) sp fromghs
                                        
    atVerbosity 3 $ "File " ++ fn ++ " read."
    case root' of
      Just root -> do
        -- imported module names
        let mods' = I.getImports root
        let mods  = map makecall mods' \\ known
        let new_known = mods ++ known
        atVerbosity 3 $ "make: branching from " ++ frommod
        indir <- mapM (determineDependencies False opts new_known) mods
        (dir,fc) <- checkCompilation sp frommod
        -- all dependencies done; we can compile (if necessary)
        if   or indir || dir
         then case fc of
                Just (_,fromghs,fromghi,fromhs) ->
                          do atVerbosity 0 $ "make: compiling " ++ fromghs
                             compile opts fromghs fromghi fromhs
                             return True
                Nothing ->
                          do atVerbosity 3 $ "make: no .ghs file for " ++ frommod
                             return False {- change here for Haskell compilation -}
         else do atVerbosity 3 $ "make: " ++ frommod ++ " unchanged"
                 return False
      _ -> return False
  where makecall m               =  (m,m ++ ".ghs",m ++ ".ghi",m ++ ".hs")

\end{code}

The |compile| function directs the compilation process.
    
\begin{code}

compile                         :: [Flag] -> String -> String -> String 
                                -> IO ()
compile opts modInName 
             modIName
             modOutName         =  do 
    -- verbosity currently depends on the number of times that
    -- the verbosity option is given on the command line
    let verbosity               =  length (filter (==Verbose) opts)
    -- the search path is the collection of all paths given in
    -- library options
    sp                          <- expandPath $
                                   foldr (\o r -> 
                                     case o of
                                       Path p -> p : r
                                       _      -> r) 
                                         searchPath opts
    -- the maximum iteration number defaults to 50, but can be set
    -- as a command line option
    let max_it                  =  head $
                                   concatMap (\o -> case o of { Cut n -> [n] ; _ -> [] })
                                             (opts ++ [Cut 50])
    -- check if we continue after errors
    let continue                =  not . null $ filter (==Continue) opts
    -- atVerbosity helps us to show certain output only if the
    -- verbosity level is high enough
    let atVerbosity n s         =  when (verbosity >= n) $ putStrLn s 

    -- Parse input file
    (root',fn)                  <- GHP.parseFile (verbosity > 0) [] modInName
    atVerbosity 1 $ "File " ++ fn ++ " read."
    case root' of 
      Just root -> 
       do 
          -- Process imports
          root <- I.processImportList root verbosity sp
          -- Resolve names
          (errs,stop,root,eV,eT,iE,hE,mods) <- return $ NR.resolve root verbosity
          -- Print errors
          putStr errs
          atVerbosity 2 $ "Name resolving complete."
          -- End program if there were errors
          if stop && not continue then exitFailure else do
          -- Analyse program
          (errs,stop) <- return $ A.sem_Root root verbosity
          -- Print errors
          putStr errs
          atVerbosity 2 $ "Static analysis complete."
          -- End program if there were errors
          if stop && not continue then exitFailure else do
          {-
          -- Write output file header
          writeFile modOutName (unlines (preamble modOutName))
          atVerbosity 3 $ "Written output file header."
          -}
          -- call compiler
          (errors,stop,screen,file,ifile) <- return $ GH.sem_Root root (verbosity,max_it,eV,eT,iE,hE,mods)
          -- Print errors
          putStr errors
          -- End program if necessary
          if stop && not continue then exitFailure else do
          -- Print screen output
          putStr screen
          -- write interface file
          writeFile modIName ifile
          atVerbosity 1 $ "Interface file " ++ modIName ++ " written."
          -- write output file
          writeFile modOutName file
          atVerbosity 1 $ "Haskell file " ++ modOutName ++ " written."
          {-
          -- pretty print input file to output
          let (Root_Root m) = root
            in  writeFile modOutName (concat . map (\x -> disp x 80 "\n") $ (PP.ppModule m))
          -}
      _ -> {- we had parse errors -}
           exitFailure
\end{code}

\begin{deprecated}
preamble                        :: String -> [String]
preamble modOutName             =  [ "module " ++ (basename modOutName) 
                                               ++ " where"
                                   , "import GHPrelude"
                                   ]
\end{deprecated}

%---------------------------------------------------------------------------
\subsection{The main function}
%---------------------------------------------------------------------------

The |main| function retrieves the input file name from the
command line or asks for it, then it calls the compiler.

\begin{code}

main                            :: IO ()
main                            =  do
                                     -- we introduce ourselves 
                                     -- putStrLn signon
                                     -- we get the command line
                                     args <- getArgs 
                                     -- we get the program name
                                     progName <- getProgName
                                     -- give the work to another function
                                     mainWithNameAndArgs progName args

mainWithNameAndArgs             :: String -> [String] -> IO ()  
mainWithNameAndArgs progName 
                    args        =  do
                                     -- analyse command line options
                                     let opts = selectOptions progName args
                                     case opts of
                                       -- we had errors
                                       Right s      -> do putStrLn s
                                       -- no file specified
                                       Left  (o,[]) -> interactive (Verbose:o)
                                       Left  (o,n ) -> nonInteractive o n

\end{code}

Interactive mode: Ask for the Generic Haskell program file.

\begin{code}

interactive                     :: [Flag] -> IO ()
interactive opts                =  
    do
      t <- if (Version `elem` opts) then displayVersionInfo >> return True
                                    else return False
      t <- if (Help    `elem` opts) then displayUsageInfo >> return True
                                    else return t
      -- putStrLn $ "Options are: " ++ show opts
      if    not t 
       then do displayVersionInfo
               putStr   $ "Enter filename: "
               hFlush stdout
               name <- getLineWithBackspaces
               -- call the compiler on the given file
               findAndCompile opts name
       else return ()

\end{code}

NonInteractive mode: Process the given program files.

\begin{code}

nonInteractive                  :: [Flag] -> [String] -> IO ()     
nonInteractive opts names       = 
    do 
      -- putStrLn $ "Options are: " ++ show opts
      -- process the files
      sequence_ (map (findAndCompile opts) names)

\end{code}

Functions displaying version and program usage info.

\begin{code}

displayVersionInfo              :: IO ()
displayVersionInfo              =  putStrLn signon

displayUsageInfo                :: IO ()
displayUsageInfo                =  putStr $ usageInfo uheader options

signon,version                  :: String        
signon                          =  "Generic Haskell compiler, version "
                                ++ version
version                         =  "1.23 (Beryl release), " ++ operatingSystem ++ " build"

{- 
  color release possibilities:
    Amber
    Blue, Brown, Beryl
    Coral, Crimson, Cyan, Citrine
    Diamond (?)
    Emerald
-}

uheader                         :: String
uheader                         =  "Usage: gh [options...] files..."

\end{code}

%---------------------------------------------------------------------------
\subsection{Option handling}
%---------------------------------------------------------------------------

There is nothing special about the options. Only thing to note: You can
specify |Verbose| as an option multiple times. The output is supposed to
be more verbose then. A general guideline is:

\begin{center}
\begin{tabular}{ll} \hline
verbosity level & output \\ \hline
0 & as quiet as possible, only errors are shown \\
1 & display warnings and errors, minor diagnostic output \\
2 & display some intermediate results \\
3+ & debug output \\ \hline
\end{tabular}
\end{center}

\begin{code}

selectOptions                   :: String 
                                -> [String] 
                                -> Either ([Flag],[String]) -- options
                                          String            -- error
selectOptions progName args     = case getOpt Permute options args of
                                    (o,n,[]  ) -> Left  $ (o,n)
                                    (_,_,errs) -> Right $ concat errs 
                                                       ++ usageInfo uheader 
                                                                    options


data Flag                       =  Help
                                |  Version 
                                |  Verbose 
                                |  Continue
                                |  Make
                                |  Output String 
                                |  Cut    Int
                                |  Path   String 
                                deriving (Show,Eq)

options                         :: [OptDescr Flag]
options                         =
    [ Option ['v']     ["verbose"]       (NoArg Verbose)      
      "(number of v's controls the verbosity)"
    , Option ['m']     ["make"]          (NoArg Make)
      "follow dependencies"
    , Option ['V'] ["version","release"] (NoArg Version)
      "show version info"
    , Option ['h','?'] ["help"]          (NoArg Help)         
      "show help"
    , Option ['c']     ["cut"]           (ReqArg (Cut . read) "N")
      "cut computations after N iterations"
    , Option ['C']     ["continue"]      (NoArg Continue)
      "continue after errors"
    , Option ['L']     ["library"]       (ReqArg Path "DIR")
      "add DIR to search path"
 -- , Option ['o']     ["output"]        (OptArg out "FILE")  
 --   "place output in FILE"
    ]

\end{code}
